/*
The following licence has been copied from
http://www.opensource.org/licenses/mit-license.html

Copyright (c) 2000, Peter Arthur Loeb

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package com.palserv.XMidi;

import org.w3c.dom.*;
import java.io.*;

/**
 * Convert an XMidi file (XML) to a midi file.
 * <p>
 * This class can be used in two ways:
 * <ol>
 * <li>
 * From the command line.
 * <li>
 * From another java class:
 * <ol>
 * <li>
 * Create an array of Strings
 * which will correspond to the
 * command line arguments.
 * <li>
 * Pass the array to the constructor:
 * <pre>
 * XM xm = new XM(args);
 * </pre>
 * <li>
 * Invoke the main conversion method:
 * <pre>
 * xm.xMidiToMidi();
 * </pre>
 * or invoke one of the two sub-methods:
 * <pre>
 * Document xDoc = xMidiToDoc(&quot;blah.xmi&quot;);
 * </pre>
 * or
 * <pre>
 * docToMidi(xDoc, &quot;blah.mid&quot;);
 * </pre>
 * </ol>
 * </ol>
 * <p>
 * Even though the XMidi format has a DTD (XMidi.dtd)
 * associated with it, this program does quite a bit of
 * its own validation.	This may be somewhat redundant,
 * but it will allow the program to work on an XMidi file
 * which does not name a DTD.  The validation forces many
 * of the same things as the DTD, although there are differences.
 * @author Peter Arthur Loeb
*/
public class XM {
	private String inFile;
	private String outFile;
	private int midiFormat;
	private int ppnq;
	private int numTracks;
	private boolean chunkFlag = false;
	private String chunkType;
	private int chunkLen;
	private FileOutputStream fos;
	private String parserName;
	private String elementName;
	private Util util = new Util();
	private String sval;
	private int slen;
	private boolean edFlag;
	private String debugFile;
	private FileWriter deb;

/**
 * public constructor
 *
 * Takes &quot;command line&quot; arguments.
 * see the
 * {@link com.palserv.XMidi.Util#getCmdLine(String[]) Util.getCmdLine}
 * method for details.
 * Also, see the main package description
 * for details about external parsers.
 * @param String[] args - command line arguments (as in main)
*/
	public XM(String[] args) {
		init();
		CmdLineArgs cla = util.getCmdLine(args);
		parserName = cla.parserName;
		inFile = cla.inFile;
		outFile = cla.outFile;
		debugFile = cla.debugFile;
	}

	void init() {
		chunkFlag = false;
		parserName = "XSilfide";
		edFlag = false;
		debugFile = null;
		deb = null;
	}

/**
 * Convert an XMidi file (XML) to a midi file.
 * Uses the two &quot;sub-methods&quot; xMidiToDoc
 * and docToMidi
*/
	public void xMidiToMidi() {

		Document xDoc = xMidiToDoc(inFile);
		docToMidi(xDoc, outFile);
	}

/**
 * Read in a XMidi file (XML) and produce a DOM structure.
 * Use an external parser to do this.
 * See the main package description
 * for details about external parsers.
 * @param String fid - name of input (XMidi) file.
 * @return Document
*/
	public Document xMidiToDoc(String fid) {
		if (debugFile != null) {
			try {
				deb = new FileWriter(debugFile);
			} catch(IOException e) {
				util.ept(e,"Unable to open "+debugFile);
			}
		}

		Document xDoc = util.getDoc(inFile, true, parserName);
		return xDoc;
	}

/**
 * Traverse a DOM structure and produce a MIDI file.
 * @param Document xDoc - the DOM Document for the DOM structure
 * @param String outFile - the file name of the output (MIDI) file.
*/
	public void docToMidi(Document xDoc, String outFile) {
		try {
			fos = new FileOutputStream(outFile);
		} catch(FileNotFoundException e) {
			util.ept(e,"unable to open "+outFile);
		}

		// write midi file
		writeMidiFile(xDoc);

		try {
			fos.close();
			if (deb != null) {
				deb.close();
			}
		} catch(IOException e) {
			util.ept(e,"error closing "+outFile);
		}
	}

	void writeMidiFile(Document xDoc) {
		Element root = (Element)xDoc.getDocumentElement();
		if (!root.getNodeName().equals("XMidi")) {
			util.rpt("root not XMidi");
		}

		String ver = root.getAttribute("VERSION");
		if (ver == null) {
			util.rpt("no version on root!");
		}

		if (!ver.equals(Version._VERSION)) {
			util.rpt("wrong version; expecting "+Version._VERSION+", got "+ver);
		}
		doElem(root);
	}

	void doElem(Node xe) {
		processElement(xe);
		if (!xe.hasChildNodes()) {
			return;
		}
		NodeList xnl = (NodeList)xe.getChildNodes();
		if (xnl == null) {
			return;
		}
		int len = xnl.getLength();
		for (int i = 0; i < len; i ++) {
			Node nx = (Node)xnl.item(i);
			doElem(nx);
		}
		if (xe.getNodeName().equals("CHUNK")) {
			chunkFlag = false;
		}
	}

	void processElement(Node xe) {
		String parent = xe.getParentNode().getNodeName();
		if (xe.getNodeType() == Node.ELEMENT_NODE) {
			Element elem = (Element)xe;
			elementName = elem.getTagName();
			if (elementName.equals("CHUNK")) {
	spt("chunk value = "+xe.getNodeValue());
				chunkFlag = true;
				chunkType = elem.getAttribute("TYPE");
				writeString(chunkType, 4);
				String ilen = elem.getAttribute("LENGTH");
				Integer iLen = new Integer(ilen);
				chunkLen = iLen.intValue();
				writeFullWord(chunkLen);
	spt("chunk type is: "+chunkType+" with length "+ilen);
			} else if (elementName.equals("MThd")) {
				chunkFlag = true;
				chunkType = elem.getAttribute("TYPE");
				writeString(chunkType, 4);
				String ilen = elem.getAttribute("LENGTH");
				chunkLen = stringToInt(ilen);
				writeFullWord(chunkLen);
			} else if (elementName.equals("MTrk")) {
				chunkFlag = true;
				chunkType = elem.getAttribute("TYPE");
				writeString(chunkType, 4);
				String ilen = elem.getAttribute("LENGTH");
				chunkLen = stringToInt(ilen);
				writeFullWord(chunkLen);
spt("we have MTrk chunk; len = "+chunkLen);
			} else if (elementName.equals("DELTA")) {
				if (!chunkFlag) {
					util.rpt("DELTA element only valid within a chunk");
				}
				if (!chunkType.equals("MTrk")) {
					util.rpt("DELTA element must be in MTrk chunk");
				}
				if (!parent.equals("MTrk")) {
					util.rpt("DELTA element must be child of MTrk");
				}
				String dtime = elem.getAttribute("DTIME");
				byte[] hdt = util.hexToBa(dtime,4);
				byte[] dt = intToDelta(hdt);
spt("Delta: "+dtime+", out = "+util.baToHex(dt,0,dt.length - 1));
				wpt(dt);
			} else if (elementName.equals("STATUS")) {
				if (!chunkFlag) {
					util.rpt("STATUS element only valid within a chunk");
				}
				if (!chunkType.equals("MTrk")) {
					util.rpt("STATUS element must be in MTrk chunk");
				}
				if (!parent.equals("DELTA")) {
					util.rpt("STATUS element must be child of DELTA");
				}
				sval = elem.getAttribute("SVAL");
				writeHex(sval,1);
				String sl = elem.getAttribute("SLEN");
				slen = stringToInt(sl);
spt("Status: "+sval+", len = "+slen);
				if (sval.equals("FF")) {
					String nmd = elem.getAttribute("SNMT");
					writeHex(nmd,1);
					byte[] hdt = util.intToBa(slen,4);
					byte[] xdt = intToDelta(hdt);
					wpt(xdt);
					if (slen == 0) {
						edFlag = true;
					}
spt("Non-midi: "+nmd);
				} else if (sval.equals("F0")) {
					byte[] hdt = util.intToBa(stringToInt(sl),4);
					wpt(intToDelta(hdt));
spt("Sysex");
				} else if (sval.equals("F6") | sval.equals("F8") |
						sval.equals("FA") | sval.equals("FB") |
						sval.equals("FC") | sval.equals("FE")) {
					edFlag = true;
spt("no data");
				}
			} else if (elementName.equals("EDATA")) {
				if (!chunkFlag) {
					util.rpt("EDATA element only valid within a chunk");
				}
				if (!chunkType.equals("MTrk")) {
					util.rpt("EDATA element must be in MTrk chunk");
				}
				if (!parent.equals("DELTA") &&
						!parent.equals("STATUS")) {
					util.rpt("STATUS element must be child of " +
						"DELTA or STATUS");
				}
spt("EDATA (element, not text)");
			} else if (elementName.equals("CHANNEL")) {
				if (!chunkFlag) {
					util.rpt("CHANNEL element only valid within a chunk");
				}
				if (!chunkType.equals("MTrk")) {
					util.rpt("CHANNEL element must be in MTrk chunk");
				}
				if (!parent.equals("DELTA") &&
						!parent.equals("STATUS")) {
					util.rpt("STATUS element must be child of " +
						"DELTA or STATUS");
				}
spt("Channel");
			} else if (elementName.equals("NOTE_OFF")) {
				if (!chunkFlag) {
					util.rpt("NOTE_OFF element only valid within a chunk");
				}
				if (!chunkType.equals("MTrk")) {
					util.rpt("NOTE_OFF element must be in MTrk chunk");
				}
				if (!parent.equals("CHANNEL")) {
					util.rpt("NOTE_OFF element must be child of CHANNEL");
				}
				String pitch = elem.getAttribute("PITCH");
				wpt(stringToInt(pitch));
				String vel = elem.getAttribute("VELOCITY");
				wpt(stringToInt(vel));
spt("Note off - "+pitch+", "+vel);
			} else if (elementName.equals("NOTE_ON")) {
				if (!chunkFlag) {
					util.rpt("NOTE_ON element only valid within a chunk");
				}
				if (!chunkType.equals("MTrk")) {
					util.rpt("NOTE_ON element must be in MTrk chunk");
				}
				if (!parent.equals("CHANNEL")) {
					util.rpt("NOTE_ON element must be child of CHANNEL");
				}
				String pitch = elem.getAttribute("PITCH");
				wpt(stringToInt(pitch));
				String vel = elem.getAttribute("VELOCITY");
				wpt(stringToInt(vel));
spt("Note on - "+pitch+", "+vel);
			} else if (elementName.equals("AFTER")) {
				if (!chunkFlag) {
					util.rpt("AFTER element only valid within a chunk");
				}
				if (!chunkType.equals("MTrk")) {
					util.rpt("AFTER element must be in MTrk chunk");
				}
				if (!parent.equals("CHANNEL")) {
					util.rpt("AFTER element must be child of CHANNEL");
				}
				String pitch = elem.getAttribute("PITCH");
				wpt(stringToInt(pitch));
				String pres = elem.getAttribute("PRESSURE");
				wpt(stringToInt(pres));
spt("AFTER - "+pitch+", "+pres);
			} else if (elementName.equals("CONTROL")) {
				if (!chunkFlag) {
					util.rpt("CONTROL element only valid within a chunk");
				}
				if (!chunkType.equals("MTrk")) {
					util.rpt("CONTROL element must be in MTrk chunk");
				}
				if (!parent.equals("CHANNEL")) {
					util.rpt("CONTROL element must be child of CHANNEL");
				}
				String cnum = elem.getAttribute("NUMBER");
				wpt(stringToInt(cnum));
				String val = elem.getAttribute("VALUE");
				wpt(stringToInt(val));
spt("CONTROL - "+cnum+", "+val);
			} else if (elementName.equals("PROGRAM")) {
				if (!chunkFlag) {
					util.rpt("PROGRAM element only valid within a chunk");
				}
				if (!chunkType.equals("MTrk")) {
					util.rpt("PROGRAM element must be in MTrk chunk");
				}
				if (!parent.equals("CHANNEL")) {
					util.rpt("PROGRAM element must be child of CHANNEL");
				}
				String patch = elem.getAttribute("NUMBER");
				wpt(stringToInt(patch));
spt("PATCH - "+patch);
			} else if (elementName.equals("PRESSURE")) {
				if (!chunkFlag) {
					util.rpt("PRESSURE element only valid within a chunk");
				}
				if (!chunkType.equals("MTrk")) {
					util.rpt("PRESSURE element must be in MTrk chunk");
				}
				if (!parent.equals("CHANNEL")) {
					util.rpt("PRESSURE element must be child of CHANNEL");
				}
				String amt = elem.getAttribute("AMOUNT");
				wpt(stringToInt(amt));
spt("PRESSURE - "+amt);
			} else if (elementName.equals("WHEEL")) {
				if (!chunkFlag) {
					util.rpt("WHEEL element only valid within a chunk");
				}
				if (!chunkType.equals("MTrk")) {
					util.rpt("WHEEL element must be in MTrk chunk");
				}
				if (!parent.equals("CHANNEL")) {
					util.rpt("WHEEL element must be child of CHANNEL");
				}
				String amt = elem.getAttribute("AMOUNT");
				int a = stringToInt(amt);
				int b = a;
				int c = a;
				b &= 127;
				c >>= 7;
				wpt(c);
				wpt(b);
spt("Wheel - "+a+": ("+c+","+a+")");
			}
		}  // end if node type is element

		if (xe.getNodeType() == Node.TEXT_NODE) {

			if (parent.equals("FORMAT")) {
				if (!chunkFlag) {
					util.rpt("FORMAT element only valid within a chunk");
				}
				if (!chunkType.equals("MThd")) {
					util.rpt("FORMAT element must be in MThd chunk");
				}
				String typ = xe.getNodeValue();
				if (typ == null) {
					util.rpt("null value for MThd->FORMAT");
				}
				typ = typ.trim();
				if (typ == null) {
					util.rpt("null value for MThd->FORMAT (after trim)");
				}
				for (int i = 0; i < typ.length(); i ++) {
					if (typ.substring(i,i+1).compareTo("0") < 0 |
							typ.substring(i,i+1).compareTo("9") > 0) {
						util.rpt("invalid numeric midi format: "+typ);
					}
				}
spt("chunk->format is "+typ+" (parent is "+parent+")");
				midiFormat = stringToInt(typ);
				writeHalfWord(midiFormat);
			}

			if (parent.equals("TRACKS")) {
				if (!chunkFlag) {
					util.rpt("TRACKS element only valid within a chunk");
				}
				if (!chunkType.equals("MThd")) {
					util.rpt("TRACKS element must be in MThd chunk");
				}
				String sNum = xe.getNodeValue();
				if (sNum == null) {
					util.rpt("null value for MThd->TRACKS");
				}
				sNum = sNum.trim();
				if (sNum == null) {
					util.rpt("null value for MThd->TRACKS (after trim)");
				}
spt("chunk->tracks is "+sNum);
				Integer iNum = new Integer(sNum);
				numTracks = iNum.intValue();
				writeHalfWord(numTracks);
			}

			if (parent.equals("PPNQ")) {
				if (!chunkFlag) {
					util.rpt("PPNQ element only valid within a chunk");
				}
				if (!chunkType.equals("MThd")) {
					util.rpt("PPNQ element must be in MThd chunk");
				}
				String pnq = xe.getNodeValue();
				if (pnq == null) {
					util.rpt("null value for MThd->PPNQ");
				}
				pnq = pnq.trim();
				if (pnq == null) {
					util.rpt("null value for MThd->PPNQ (after trim)");
				}
spt("chunk->ppnq is "+pnq);
	/*
				Integer iTyp = new Integer(xe.getNodeValue());
				ppnq = iTyp.intValue();
	*/
				writeHex(pnq,2);
			}

			if (parent.equals("HEXDATA")) {
				if (!chunkFlag) {
					util.rpt("HEXDATA element only valid within CHUNK");
				}
				String xdata = xe.getNodeValue();
				if (xdata == null) {
					util.rpt("null value for CHUNK->HEXDATA");
				}
				writeHex(xdata,chunkLen);
			}

			if (parent.equals("EDATA")) {
				if (!chunkFlag) {
					util.rpt("HEXDATA element only valid within CHUNK");
				}
				if (edFlag) {
					edFlag = false;
				} else {
					String xdata = xe.getNodeValue();
spt("EDATA: "+xdata);
					if (xdata == null) {
						util.rpt("null value for DELTA/STATUS->EDATA");
					}
					writeHex(xdata,slen);
				}
			}
		}
	}
/**
convert byte array to delta time
a delta time is expressed as a
byte array, length unknown (4 or less)
*/
	byte[] intToDelta(byte[] t) {
/*
from fullword binary value to variable (midi) value:
- split number into 5 bit groups of 4 7 7 7 7
- just deal with the four groups of 7 (ignore the 4)
- put each group of 7 into a byte
- case 1: whole full word is zero.
	return one byte of zero.
- case 2: some non-zero bit (after first four bits)
	- work from left to right
	- throw away bytes which are zero until non-zero byte
		encountered
	- turn on hi-ord bit on all but last byte
*/
		int i1 = util.baToInt(t,0,t.length - 1);
spt("intToDelta: i1 = "+i1+", t = "+util.baToHex(t,0,t.length - 1));
		if (i1 == 0) {
			byte[] b = new byte[1];
			b[0] = 0;
			return b;
		}
		int i2 = i1;
		byte[] b1 = new byte[4];
		for (int i = 0; i < 4; i ++) {
			int j = 4 - i - 1;
			int k = i2 % 128;
			i2 /= 128;
			b1[j] = (byte)k;
		}
		int j = -1;
		boolean bFlag = false;
		for (int i = 0; i < 4; i ++) {
			if (bFlag) { continue; }
			if (b1[i] != 0) {
				j = i;
				bFlag = true;
			}
		}
		int k = 4 - j;
		byte[] b2 = new byte[k];
		for (int i = 0; i < k; i ++) {
			b2[i] = b1[i + j];
			if (i != (k - 1)) {
				b2[i] |= 128;
			}
		}
		return b2;
	}

	int stringToInt(String s) {
		Integer i = new Integer(s);
		int r = i.intValue();
		return r;
	}

/**
write a line to standard output.
This is mostly for debugging.
If the -d command line option
is not coded, then this routine
will do nothing.
*/
	void spt(String s) {
//		  System.out.println(s);
		if (deb != null) {
			try {
				deb.write(s+"\n");
			} catch(IOException e) {
				util.ept(e,"error in debug write");
			}
		}
	}

	void prt(String s) {
		System.out.println(s);
	}

/**
 * This method is for command line invocation.
 * See the
 * {@link com.palserv.XMidi.Util#getCmdLine(String[]) Util.getCmdLine}
 * method for details of the command line syntax.
 * @param String[] args - command line
*/
	public static void main(String[] args) {
		XM xm = new XM(args);
		xm.xMidiToMidi();
	}

	void writeString(String s, int len) {
		int l = s.length();
		if (l != len) {
			util.rpt("writeString; string length ("+l+") != expected length ("+len+")");
		}
		wpt(s.getBytes());
	}

	void writeHex(String h, int len) {
		int l = h.length();
		int cnt = 0;
		int bs = 0;
		int bc = 0;
		for (int i = 0; i < l; i ++) {
			String s = h.substring(i,i+1);
			int x = " \n\t\r".indexOf(s);
			if (x != -1) {
				continue;
			}
			int tmp = "0123456789ABCDEF".indexOf(s.toUpperCase());
			if (bc == 0) {
				bs = tmp;
			} else if (bc == 1) {
				bs <<= 4;
				bs |= tmp;
			} else {
				util.rpt("writeHex; internal error");
			}
			bc ++;
			if (bc >= 2) {
				wpt(bs);
				cnt ++;
				bc = 0;
			}
		}
		if (bc != 0) {
			util.rpt("un-even number of hex digits");
		}
/*
spt("writeHex: writing outside loop: "+cnt);
		wpt(bs);
		cnt ++;
*/
		if (cnt != len) {
			util.rpt("writeHex count ("+cnt+") != length ("+len+")");
		}
	}

	void writeFullWord(int f) {
		byte[] b = util.intToBa(f,4);
		wpt(b);
	}

	void writeHalfWord(int h) {
		byte[] b = util.intToBa(h,2);
		wpt(b);
	}

	void wpt(byte[] b) {
		try {
		   fos.write(b);
		} catch(IOException e) {
			util.ept(e,"Write error");
		}
	}

/**
overloaded wpt
*/
	void wpt(int b) {
		try {
		   fos.write(b);
		} catch(IOException e) {
			util.ept(e,"Write error");
		}
	}
}
