/*
The following licence has been copied from
http://www.opensource.org/licenses/mit-license.html

Copyright (c) 2000, Peter Arthur Loeb

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

package com.palserv.XMidi;

import java.io.*;
import java.util.*;
import org.w3c.dom.*;

/**
 * Convert a midi file to an XMidi (XML) file.
 * <p>
 * This class can be used in two ways:
 * <ol>
 * <li>
 * From the command line.
 * <li>
 * From another java class:
 * <ol>
 * <li>
 * Create an array of Strings
 * which will correspond to the
 * command line arguments.
 * <li>
 * Pass the array to the constructor:
 * <pre>
 * MX mx = new MX(args);
 * </pre>
 * <li>
 * Invoke the conversion method:
 * <pre>
 * mx.midiToXMidi();
 * </pre>
 * </ol>
 * </ol>
 * @author Peter Arthur Loeb
*/
public class MX {
	private String inFile;
	private String outFile;
//	  private int midiFormat;
//	  private int ppnq;
//	  private Util util = new Util();
	private Hashtable<String,String> ffHash;
	private String[] chanArray;
	private String[] fArray;
//	  private FileWriter fw;
//	  private FileInputStream fis;
//	  private int numTracks;
//	  private int trkCount;
//	  private String parserName = "com.palserv.XMidi.JAXP";
//	  private String debugFile;
//	  private FileWriter deb;
	private boolean verbose;
	private String[] notes;
	private int[] register;
	private Hashtable<String,String> contHash;

/**
 * public constructor
 *
 * Takes &quot;command line&quot; arguments.
 * see the
 * {@link com.palserv.XMidi.Util#getCmdLine(String[]) Util.getCmdLine}
 * method for details.
 * Also, see the main package description
 * for details about external parsers.
 * @param String[] args - command line arguments (as in main)
*/
	public MX(String[] args) {
		CmdLineArgs cla = Util.getCmdLine(args);
//		  parserName = cla.parserName;
		inFile = cla.inFile;
		outFile = cla.outFile;
//		  debugFile = cla.debugFile;
		verbose = cla.verbose;
		init();
		varInit();
		if (verbose) {
			verboseInit();
		}
	}


	void varInit() {
		State state = State.getState();
		state.pw = null;
		state.fis = null;
		state.numTracks = 0;
		state.trkCount = 0;
	}
/**
 * convert a midi file to an XMidi (XML) file
*/
	public void midiToXMidi() {
		varInit();
		State state = State.getState();
		try {
			state.fis = new FileInputStream(inFile);
		} catch(FileNotFoundException e) {
			Util.ept(e,"Unable to open "+inFile);
		}
//		  Document xDoc = midiToDoc(inFile);
//		  docToXMidi(xDoc, outFile);

		try {
			state.pw = new PrintWriter(new BufferedWriter(new FileWriter(outFile)));
		} catch(IOException e) {
			Util.ept(e,"Unable to open "+outFile);
		}
/* processing goes here
prior versions converted the MIDI file to a DOM Document,
then converted the Document to an XML file.
This version eliminates the DOM Document and
goes directly from the MIDI file to XML.
*/
		doHeader(state);
		doDoc(state);
// termination
		state.pw.close();

		try {
			state.fis.close();
		} catch(IOException e) {
			Util.ept(e,"Unable to close "+inFile);
		}
	}

	void doHeader(State state) {
		// write the header
		Util.ptf("<?xml version=\"1.0\" standalone=\"no\"?>\n");
		Util.ptf("<!DOCTYPE XMidi SYSTEM \"XMidi.dtd\">\n");
		Util.ptf("<!-- produced by com.palserv.XMidi.MX from "+inFile+" -->\n");
	}

	void doDoc(State state) {
		state.indent = 0;
		Tag xm = Tag.getTag("XMidi");
		Util.ptf(xm.createStartXML());
		state.indent ++;
		boolean cflag = true;
		while (cflag) { 	// for each chunk
			byte[] hdr = new byte[8];
			try {
				int r = state.fis.read(hdr);
				if (r == -1) {
					cflag = false;
					continue;
				}
				if (r < 8) {
					Util.rpt("getting header");
				}
			} catch(IOException e) {
				Util.ept(e,"IO exception getting header");
			}
			// get chunk id
			String cid = baToString(hdr,0,3);

			// get chunk length
			state.chunkLen = Util.baToInt(hdr,4,7);

			// get rest of chunk
			state.dta = new byte[state.chunkLen];
			try {
				int r = state.fis.read(state.dta);
				if (r < state.chunkLen) {
					Util.rpt("getting data");
				}
			} catch(IOException e) {
				Util.ept(e,"IO exception getting data");
			}

			Tag mm = null;
			if (cid.equals("MThd") || cid.equals("MTrk")) {
				mm = Tag.getTag(cid);
			} else {
				mm = Tag.getTag("CHUNK");
			}
			Util.ptf(mm.createStartXML());
			if (cid.equals("MTrk")) {
				state.indent ++;
				doTrack(state);
				state.indent --;
			}
			Util.ptf(mm.createEndXML());
		}	// end of for each chunk
//		  spt("Finished");
		state.indent --;
		Util.ptf(xm.createEndXML());
	}

/**
 * main routine, executed from command line.
 * See the
 * {@link com.palserv.XMidi.Util#getCmdLine(String[]) Util.getCmdLine}
 * method for details of the command line arguments
*/
	public static void main(String[] args) {
		MX mx = new MX(args);
		mx.midiToXMidi();
	}

/**
do the track data (within the MTrk tag)*/
	void doTrack(State state) {
		boolean tFlag = true;
		int offset = 0;

		// initialize variables
		int dlen = 0;
		state.edata = null;
		state.snam = null;
		state.nmData = "";
		state.ctag = null;
		state.ctagAttr = new String[4];
		state.ctagAttrVal= new String[4];
		state.ctagnum = 0;
		state.cnum = null;
		state.ctyp = null;
		state.slen = 0;
		int chanType = 0;
		int hiName = 0;
		int status = 0;
		int noff = 0;
		state.sval = null;
		boolean ecFlag = true; // assume state.edata
		boolean nFlag = true;  // use state.slen for noff
		boolean firstTime = true;

		while (tFlag) {
			// do delta
			ByteLen bl = deltaToInt(state.dta, offset);
			offset += bl.len;
			state.deltaLen = Util.baToHex(bl.ba,0,3);

			nFlag = true;	// assume simple (state.slen) offset
			// may or may not be status
			boolean statFlag = false;
			int first = Util.baToInt(state.dta,offset,offset);
// spt("doTrack: in loop; deltaLen="+deltaLen+", len="+bl.len+", first="+first);
			if ((first & 128) == 128) {
				// it is a status byte
				statFlag = true;
				state.sval = Util.baToHex(state.dta,offset,offset);
				status = first;
// spt("doTrack: have status: "+state.sval);
				if (status < 240 && status > 127) {
					ecFlag = false;
					chanType = (status - 128) / 16;
					state.snam = chanArray[chanType];
					state.ctyp = state.sval.substring(0,1);
					state.cnum = state.sval.substring(1,2);
				} else {
					ecFlag = true;
					if (status > 239 && status < 256) {
						hiName = status - 240;
						state.snam = fArray[hiName];
					} else {
						Util.rpt("Invalid status: "+status);
					}
				}
				offset ++;
			} else {
// spt("doTrack: running status");
			}

			state.nmData = "";
			if (firstTime) {
				firstTime = false;
				if (!statFlag) {
					Util.rpt("first time, but no status; first = "+first);
				}
			}

			// offset points to the byte after the status
			// or first byte of "running status"
			Tag delta = Tag.getTag("DELTA");
			Util.ptf(delta.createStartXML());
			if (status > 127 && status < 144) {
				// note off
				state.slen = 2;
				// set up tag
				int pitch = Util.baToInt(state.dta,offset,offset);
				int vel = Util.baToInt(state.dta,offset+1,offset+1);
				state.ctag = "NOTE_OFF";
				state.ctagAttr[0] = "PITCH";
				state.ctagAttrVal[0] = ""+pitch;
				state.ctagAttr[1] = "VELOCITY";
				state.ctagAttrVal[1] = ""+vel;
				state.ctagnum = 2;
				if (verbose) {
					state.ctagAttr[2] = "NAME";
					state.ctagAttrVal[2] = notes[pitch];
					state.ctagAttr[3] = "REGISTER";
					state.ctagAttrVal[3] = ""+register[pitch];
					state.ctagnum = 4;
				}
			} else if (status > 143 && status < 160) {
				// note on
				state.slen = 2;
				int pitch = Util.baToInt(state.dta,offset,offset);
				int vel = Util.baToInt(state.dta,offset+1,offset+1);
				state.ctag = "NOTE_ON";
				state.ctagAttr[0] = "PITCH";
				state.ctagAttrVal[0] = ""+pitch;
				state.ctagAttr[1] = "VELOCITY";
				state.ctagAttrVal[1] = ""+vel;
				state.ctagnum = 2;
				if (verbose) {
					state.ctagAttr[2] = "NAME";
					state.ctagAttrVal[2] = notes[pitch];
					state.ctagAttr[3] = "REGISTER";
					state.ctagAttrVal[3] = ""+register[pitch];
					state.ctagnum = 4;
				}
			} else if (status > 159 && status < 176) {
				// after touch
				state.slen = 2;
				int pitch = Util.baToInt(state.dta,offset,offset);
				int pres = Util.baToInt(state.dta,offset+1,offset+1);
				state.ctag = "AFTER";
				state.ctagAttr[0] = "PITCH";
				state.ctagAttrVal[0] = ""+pitch;
				state.ctagAttr[1] = "PRESSURE";
				state.ctagAttrVal[1] = ""+pres;
				state.ctagnum = 2;
				if (verbose) {
					state.ctagAttr[2] = "NAME";
					state.ctagAttrVal[2] = notes[pitch];
					state.ctagAttr[3] = "REGISTER";
					state.ctagAttrVal[3] = ""+register[pitch];
					state.ctagnum = 4;
				}
			} else if (status > 175 && status < 192) {
				// control change
				state.slen = 2;
				int contnum = Util.baToInt(state.dta,offset,offset);
				int contval = Util.baToInt(state.dta,offset+1,offset+1);
				state.ctag = "CONTROL";
				state.ctagAttr[0] = "NUMBER";
				state.ctagAttrVal[0] = ""+contnum;
				state.ctagAttr[1] = "VALUE";
				state.ctagAttrVal[1] = ""+contval;
				state.ctagnum = 2;
				if (verbose) {
					state.ctagAttr[2] = "NAME";
					state.ctagAttrVal[2] = (String)contHash.get(""+contnum);
					state.ctagnum = 3;
				}
			} else if (status > 191 && status < 208) {
				// program (patch) change
				state.slen = 1;
				int patch = Util.baToInt(state.dta,offset,offset);
				state.ctag = "PROGRAM";
				state.ctagAttr[0] = "NUMBER";
				state.ctagAttrVal[0] = ""+patch;
				state.ctagnum = 1;
			} else if (status > 207 && status < 224) {
				// channel pressure
				state.slen = 1;
				int pamt = Util.baToInt(state.dta,offset,offset);
				state.ctag = "PRESSURE";
				state.ctagAttr[0] = "AMOUNT";
				state.ctagAttrVal[0] = ""+pamt;
				state.ctagnum = 1;
			} else if (status > 223 && status < 240) {
				// pitch wheel
				state.slen = 2;
				int pwamt = getPW(offset, state);
				state.ctag = "WHEEL";
				state.ctagAttr[0] = "AMOUNT";
				state.ctagAttrVal[0] = ""+pwamt;
				state.ctagnum = 1;
			} else if (status == 240) {
				// sysex
				bl = deltaToInt(state.dta, offset);
				state.slen = Util.baToInt(bl.ba,0,3);
				noff = bl.len;
				nFlag = false;
				state.edata = Util.baToHex(state.dta,offset+noff,offset+noff+state.slen-1);
				noff += state.slen;
			} else if (status == 255) {
				// non midi (reset only in "live" midi")
				state.nmData = Util.baToHex(state.dta,offset,offset);
//				  state.nmData = "SNMT=\""+nmDta+"\" ";
				state.snam = "non-MIDI";
				String nmNam = (String)ffHash.get(state.nmData);
				if (nmNam != null) {
					state.snam += " - "+nmNam;
				} else {
					state.snam += " - Unknown";
				}
//				  int nmt = Util.baToInt(state.dta,offset+1,offset+1);
				bl = deltaToInt(state.dta, offset + 1);
				state.slen = Util.baToInt(bl.ba,0,3);
				noff = bl.len+1;
				nFlag = false;
				if (state.slen == 0) {
					state.edata = "No data";
				} else {
					state.edata = Util.baToHex(state.dta,offset+noff,
						offset+noff+state.slen-1);
					noff += state.slen;
				}
				if (state.nmData.equals("58")) { // time sig
// 0 - NN, 1 - DD, 2 - CC, 3 - BB, 4 - DDD
					state.timeSig = new int[5];
					state.timeSig[0] = Util.baToInt(state.dta, offset + 2, offset + 2);
					state.timeSig[1] = Util.baToInt(state.dta, offset + 3, offset + 3);
					state.timeSig[2] = Util.baToInt(state.dta, offset + 4, offset + 4);
					state.timeSig[3] = Util.baToInt(state.dta, offset + 5, offset + 5);
					state.timeSig[4] = (int)Math.pow(2, state.timeSig[1]);
					Tag tsig = Tag.getTag("TIME_SIG");
					state.indent ++;
					Util.ptf(tsig.createStartXML());
					state.indent --;
				}
// spt("doTrack: status FF"+state.nmData+", state.edata = "+state.edata);
			} else if (status == 241 || status == 243) {
				int tcv = (int)state.dta[offset];
				Integer tc = new Integer(tcv);
				state.edata = tc.toString();
				state.slen = 1;
			} else if (status == 242) {
				int tcv = getPW(offset, state);
				Integer tc = new Integer(tcv);
				state.edata = tc.toString();
				state.slen = 2;
			} else if (status == 246 || status == 248 ||
					status == 250 || status == 251 ||
					status == 252 || status == 254) {
				state.edata = "No data for "+state.snam;
				state.slen = 0;
			} else {	// really unknown
				int ol = getNextHiOrd(offset, state);
				state.edata = Util.baToHex(state.dta,offset+1,ol);
				ol -= offset + 1;
				state.slen = ol;
			}
			Tag eStatus = null;
			if (statFlag) {
				eStatus = Tag.getTag("STATUS");
				state.indent ++;
				Util.ptf(eStatus.createStartXML());
			}
			if (ecFlag) {
				Tag edat = Tag.getTag("EDATA");
				state.indent ++;
				Util.ptf(edat.createStartXML());
				Util.ptf(edat.createEndXML());
				state.indent --;
			} else {
				state.indent ++;
				Tag channel = Tag.getTag("CHANNEL");
				Util.ptf(channel.createStartXML());
				Util.ptf(channel.createEndXML());
				state.indent --;
			}

			if (nFlag) {
				offset += state.slen;
			} else {
				offset += noff;
			}
			if (offset >= state.chunkLen) {
				tFlag = false;
			}
			if (statFlag) {
				Util.ptf(eStatus.createEndXML()); // finish off STATUS tag
				state.indent --;
			}
			Util.ptf(delta.createEndXML());   // finish off DELTA tag
		}	// end while (tflag)
	}	// end doTrack method

	int getPW(int offset, State state) {
		int hi = Util.baToInt(state.dta,offset,offset);
		int lo = Util.baToInt(state.dta,offset+1,offset+1);
		hi <<= 7;
		lo |= hi;
		return lo;
	}

	int getNextHiOrd(int offset, State state) {
		int ol = 0;
		boolean tflag = true;
		for (int o = offset+1; o < state.dta.length - 1; o ++) {
			if (tflag) {
				int x = Util.baToInt(state.dta,o,o);
				if ((x & 128) == 128) {
					tflag = false;
					ol = o;
				}
			}
		}
		if (tflag) {
			ol = offset+state.dta.length;
		}
		return ol;
	}

/**
write a line to standard output.
*/
	void spt(String s) {
		System.out.println(s);
	}

/**
convert byte array to string (not hex)
*/
	String baToString(byte[] b, int strt, int end) {
		int l = end - strt + 1;
		char[] c = new char[l];
		int i = 0;
		for (int j = 0; j < l; j ++) {
			c[j] = (char)b[j];
		}
		return new String(c);
	}

/**
convert delta time to byte array
a delta time is expressed as a
byte array, length unknown (4 or less)
*/
	public ByteLen deltaToInt(byte[] b, int offset) {
/*
- capture up to four bytes including first with hi-ord
	bit off
- turn off hi-ord bits
- accumulate 4 groups of 7 into 28 bit number with
	hi-ord 4 bits zero.
*/
		int j = 0;
		byte[] ba = new byte[4];
		boolean jFlag = true;
		while (jFlag) {
			if ((j + offset) > b.length) {
				Util.rpt("deltaToInt: length error; j = "+j+", offset = "+offset+", b.length = "+b.length);
			}
			ba[j] = b[j + offset];
			if (ba[j] >= 0) {
				jFlag = false;
			} else {
				ba[j] &= 127;
			}
// spt("deltaToInt: j = "+j+", ba = "+Util.baToInt(ba,0,j));
			j ++;
		}
		int s = 0;
		for (int i = 0; i < j; i ++) {
			int k = j - i - 1;
			int p = Util.pow(128,i);
			int m = (int)ba[k];
			s += m * p;
// spt("deltaToInt: in loop: s = "+s+", i = "+i+", j = "+j+", k = "+k+", p = "+p+", m = "+m);
		}
// spt("deltaToInt: s = "+s);
		ByteLen bl = new ByteLen(Util.intToBa(s,4),j);
		return bl;
	}

	void init() {
		ffHash = new Hashtable<String,String>();
		ffHash.put("00", "Sequence Number");
		ffHash.put("01", "Text");
		ffHash.put("02", "Copyright");
		ffHash.put("03", "Sequence/Track Name");
		ffHash.put("04", "Instrument");
		ffHash.put("05", "Lyric");
		ffHash.put("06", "Marker");
		ffHash.put("07", "Cue Point");
		ffHash.put("20", "MIDI Channel");
		ffHash.put("21", "MIDI Port");
		ffHash.put("2F", "End of Track");
		ffHash.put("51", "Tempo");
		ffHash.put("54", "SMPTE Offset");
		ffHash.put("58", "Time Signature");
		ffHash.put("59", "Key Signature");
		ffHash.put("7F", "Proprietary Event");

		chanArray = new String[7];
		chanArray[0] = "Note Off";
		chanArray[1] = "Note On";
		chanArray[2] = "After Touch";
		chanArray[3] = "Control Change";
		chanArray[4] = "Program Change";
		chanArray[5] = "Channel Pressure";
		chanArray[6] = "Pitch Wheel";

		fArray = new String[16];
		fArray[0] = "SYSEX";
		fArray[1] = "MTC Quarter Frame Message";
		fArray[2] = "Song Position Pointer";
		fArray[3] = "Song Select";
		fArray[4] = "Undefined";
		fArray[5] = "Undefined";
		fArray[6] = "Tune Request";
		fArray[7] = "Unsupported";
		fArray[8] = "MIDI Clock";
		fArray[9] = "Undefined";
		fArray[10] = "MIDI Start";
		fArray[11] = "MIDI Continue";
		fArray[12] = "MIDI Stop";
		fArray[13] = "Undefined";
		fArray[14] = "Active Sense";
		fArray[15] = "NotUnderstood";
	}

	void verboseInit() {
		contHash = new Hashtable<String,String>();

		contHash.put("0", "Bank Select");
		contHash.put("1", "Modulation Wheel (coarse)");
		contHash.put("2", "Breath controller (coarse)");
		contHash.put("4", "Foot Pedal (coarse)");
		contHash.put("5", "Portamento Time (coarse)");
		contHash.put("6", "Data Entry (coarse)");
		contHash.put("7", "Volume (coarse)");
		contHash.put("8", "Balance (coarse)");
		contHash.put("10", "Pan position (coarse)");
		contHash.put("11", "Expression (coarse)");
		contHash.put("12", "Effect Control 1 (coarse)");
		contHash.put("13", "Effect Control 2 (coarse)");
		contHash.put("16", "General Purpose Slider 1");
		contHash.put("17", "General Purpose Slider 2");
		contHash.put("18", "General Purpose Slider 3");
		contHash.put("19", "General Purpose Slider 4");
		contHash.put("32", "Bank Select (fine)");
		contHash.put("33", "Modulation Wheel (fine)");
		contHash.put("34", "Breath controller (fine)");
		contHash.put("36", "Foot Pedal (fine)");
		contHash.put("37", "Portamento Time (fine)");
		contHash.put("38", "Data Entry (fine)");
		contHash.put("39", "Volume (fine)");
		contHash.put("40", "Balance (fine)");
		contHash.put("42", "Pan position (fine)");
		contHash.put("43", "Expression (fine)");
		contHash.put("44", "Effect Control 1 (fine)");
		contHash.put("45", "Effect Control 2 (fine)");
		contHash.put("64", "Hold Pedal (on/off)");
		contHash.put("65", "Portamento (on/off)");
		contHash.put("66", "Sustenuto Pedal (on/off)");
		contHash.put("67", "Soft Pedal (on/off)");
		contHash.put("68", "Legato Pedal (on/off)");
		contHash.put("69", "Hold 2 Pedal (on/off)");
		contHash.put("70", "Sound Variation");
		contHash.put("71", "Sound Timbre");
		contHash.put("72", "Sound Release Time");
		contHash.put("73", "Sound Attack Time");
		contHash.put("74", "Sound Brightness");
		contHash.put("75", "Sound Control 6");
		contHash.put("76", "Sound Control 7");
		contHash.put("77", "Sound Control 8");
		contHash.put("78", "Sound Control 9");
		contHash.put("79", "Sound Control 10");
		contHash.put("80", "General Purpose Button 1 (on/off)");
		contHash.put("81", "General Purpose Button 2 (on/off)");
		contHash.put("82", "General Purpose Button 3 (on/off)");
		contHash.put("83", "General Purpose Button 4 (on/off)");
		contHash.put("91", "Effects Level");
		contHash.put("92", "Tremulo Level");
		contHash.put("93", "Chorus Level");
		contHash.put("94", "Celeste Level");
		contHash.put("95", "Phaser Level");
		contHash.put("96", "Data Button increment");
		contHash.put("97", "Data Button decrement");
		contHash.put("98", "Non-registered Parameter (fine)");
		contHash.put("99", "Non-registered Parameter (coarse)");
		contHash.put("100", "Registered Parameter (fine)");
		contHash.put("101", "Registered Parameter (coarse)");
		contHash.put("120", "All Sound Off");
		contHash.put("121", "All Controllers Off");
		contHash.put("122", "Local Keyboard (on/off)");
		contHash.put("123", "All Notes Off");
		contHash.put("124", "Omni Mode Off");
		contHash.put("125", "Omni Mode On");
		contHash.put("126", "Mono Operation");
		contHash.put("127", "Poly Operation");

		notes = new String[128];
		register = new int[128];
		notes[0] = "C";
		register[0] = -5;
		notes[1] = "C#";
		register[1] = -5;
		notes[2] = "D";
		register[2] = -5;
		notes[3] = "Eb";
		register[3] = -5;
		notes[4] = "E";
		register[4] = -5;
		notes[5] = "F";
		register[5] = -5;
		notes[6] = "F#";
		register[6] = -5;
		notes[7] = "G";
		register[7] = -5;
		notes[8] = "Ab";
		register[8] = -5;
		notes[9] = "A";
		register[9] = -5;
		notes[10] = "Bb";
		register[10] = -5;
		notes[11] = "B";
		register[11] = -5;
		notes[12] = "C";
		register[12] = -4;
		notes[13] = "C#";
		register[13] = -4;
		notes[14] = "D";
		register[14] = -4;
		notes[15] = "Eb";
		register[15] = -4;
		notes[16] = "E";
		register[16] = -4;
		notes[17] = "F";
		register[17] = -4;
		notes[18] = "F#";
		register[18] = -4;
		notes[19] = "G";
		register[19] = -4;
		notes[20] = "Ab";
		register[20] = -4;
		notes[21] = "A";
		register[21] = -4;
		notes[22] = "Bb";
		register[22] = -4;
		notes[23] = "B";
		register[23] = -4;
		notes[24] = "C";
		register[24] = -3;
		notes[25] = "C#";
		register[25] = -3;
		notes[26] = "D";
		register[26] = -3;
		notes[27] = "Eb";
		register[27] = -3;
		notes[28] = "E";
		register[28] = -3;
		notes[29] = "F";
		register[29] = -3;
		notes[30] = "F#";
		register[30] = -3;
		notes[31] = "G";
		register[31] = -3;
		notes[32] = "Ab";
		register[32] = -3;
		notes[33] = "A";
		register[33] = -3;
		notes[34] = "Bb";
		register[34] = -3;
		notes[35] = "B";
		register[35] = -3;
		notes[36] = "C";
		register[36] = -2;
		notes[37] = "C#";
		register[37] = -2;
		notes[38] = "D";
		register[38] = -2;
		notes[39] = "Eb";
		register[39] = -2;
		notes[40] = "E";
		register[40] = -2;
		notes[41] = "F";
		register[41] = -2;
		notes[42] = "F#";
		register[42] = -2;
		notes[43] = "G";
		register[43] = -2;
		notes[44] = "Ab";
		register[44] = -2;
		notes[45] = "A";
		register[45] = -2;
		notes[46] = "Bb";
		register[46] = -2;
		notes[47] = "B";
		register[47] = -2;
		notes[48] = "C";
		register[48] = -1;
		notes[49] = "C#";
		register[49] = -1;
		notes[50] = "D";
		register[50] = -1;
		notes[51] = "Eb";
		register[51] = -1;
		notes[52] = "E";
		register[52] = -1;
		notes[53] = "F";
		register[53] = -1;
		notes[54] = "F#";
		register[54] = -1;
		notes[55] = "G";
		register[55] = -1;
		notes[56] = "Ab";
		register[56] = -1;
		notes[57] = "A";
		register[57] = -1;
		notes[58] = "Bb";
		register[58] = -1;
		notes[59] = "B";
		register[59] = -1;
		notes[60] = "C";
		register[60] = 0;
		notes[61] = "C#";
		register[61] = 0;
		notes[62] = "D";
		register[62] = 0;
		notes[63] = "Eb";
		register[63] = 0;
		notes[64] = "E";
		register[64] = 0;
		notes[65] = "F";
		register[65] = 0;
		notes[66] = "F#";
		register[66] = 0;
		notes[67] = "G";
		register[67] = 0;
		notes[68] = "Ab";
		register[68] = 0;
		notes[69] = "A";
		register[69] = 0;
		notes[70] = "Bb";
		register[70] = 0;
		notes[71] = "B";
		register[71] = 0;
		notes[72] = "C";
		register[72] = 1;
		notes[73] = "C#";
		register[73] = 1;
		notes[74] = "D";
		register[74] = 1;
		notes[75] = "Eb";
		register[75] = 1;
		notes[76] = "E";
		register[76] = 1;
		notes[77] = "F";
		register[77] = 1;
		notes[78] = "F#";
		register[78] = 1;
		notes[79] = "G";
		register[79] = 1;
		notes[80] = "Ab";
		register[80] = 1;
		notes[81] = "A";
		register[81] = 1;
		notes[82] = "Bb";
		register[82] = 1;
		notes[83] = "B";
		register[83] = 1;
		notes[84] = "C";
		register[84] = 2;
		notes[85] = "C#";
		register[85] = 2;
		notes[86] = "D";
		register[86] = 2;
		notes[87] = "Eb";
		register[87] = 2;
		notes[88] = "E";
		register[88] = 2;
		notes[89] = "F";
		register[89] = 2;
		notes[90] = "F#";
		register[90] = 2;
		notes[91] = "G";
		register[91] = 2;
		notes[92] = "Ab";
		register[92] = 2;
		notes[93] = "A";
		register[93] = 2;
		notes[94] = "Bb";
		register[94] = 2;
		notes[95] = "B";
		register[95] = 2;
		notes[96] = "C";
		register[96] = 3;
		notes[97] = "C#";
		register[97] = 3;
		notes[98] = "D";
		register[98] = 3;
		notes[99] = "Eb";
		register[99] = 3;
		notes[100] = "E";
		register[100] = 3;
		notes[101] = "F";
		register[101] = 3;
		notes[102] = "F#";
		register[102] = 3;
		notes[103] = "G";
		register[103] = 3;
		notes[104] = "Ab";
		register[104] = 3;
		notes[105] = "A";
		register[105] = 3;
		notes[106] = "Bb";
		register[106] = 3;
		notes[107] = "B";
		register[107] = 3;
		notes[108] = "C";
		register[108] = 4;
		notes[109] = "C#";
		register[109] = 4;
		notes[110] = "D";
		register[110] = 4;
		notes[111] = "Eb";
		register[111] = 4;
		notes[112] = "E";
		register[112] = 4;
		notes[113] = "F";
		register[113] = 4;
		notes[114] = "F#";
		register[114] = 4;
		notes[115] = "G";
		register[115] = 4;
		notes[116] = "Ab";
		register[116] = 4;
		notes[117] = "A";
		register[117] = 4;
		notes[118] = "Bb";
		register[118] = 4;
		notes[119] = "B";
		register[119] = 4;
		notes[120] = "C";
		register[120] = 5;
		notes[121] = "C#";
		register[121] = 5;
		notes[122] = "D";
		register[122] = 5;
		notes[123] = "Eb";
		register[123] = 5;
		notes[124] = "E";
		register[124] = 5;
		notes[125] = "F";
		register[125] = 5;
		notes[126] = "F#";
		register[126] = 5;
		notes[127] = "G";
		register[127] = 5;
	}
}
