/*
The following licence has been copied from
http://www.opensource.org/licenses/mit-license.html

Copyright (c) 2000, Peter Arthur Loeb

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

package com.palserv.XMidi;

import java.io.*;
import java.util.*;
import org.w3c.dom.*;

/**
 * Convert a midi file to an XMidi (XML) file.
 * <p>
 * This class can be used in two ways:
 * <ol>
 * <li>
 * From the command line.
 * <li>
 * From another java class:
 * <ol>
 * <li>
 * Create an array of Strings
 * which will correspond to the
 * command line arguments.
 * <li>
 * Pass the array to the constructor:
 * <pre>
 * MX mx = new MX(args);
 * </pre>
 * <li>
 * Invoke the conversion method:
 * <pre>
 * mx.midiToXMidi();
 * </pre>
 * or invoke one of the &quot;inner&quot; conversion
 * methods:
 * <pre>
 * Document doc = mx.midiToDoc(&quot;blah.mid&quot;);
 * </pre>
 * or
 * <pre>
 * mx.docToXMidi(doc, &quot;blah.xmi&quot;);
 * </pre>
 * </ol>
 * </ol>
 * @author Peter Arthur Loeb
*/
public class MX {
	private String inFile;
	private String outFile;
	private int midiFormat;
	private int ppnq;
	private Util util = new Util();
	private Hashtable<String,String> ffHash;
	private String[] chanArray;
	private String[] fArray;
	private FileWriter fw;
	private FileInputStream fis;
	private int numTracks;
	private int trkCount;
	private String parserName;
	private String debugFile;
	private FileWriter deb;
	private boolean verbose;
	private String[] notes;
	private int[] register;
	private Hashtable<String,String> contHash;

/**
 * public constructor
 *
 * Takes &quot;command line&quot; arguments.
 * see the
 * {@link com.palserv.XMidi.Util#getCmdLine(String[]) Util.getCmdLine}
 * method for details.
 * Also, see the main package description
 * for details about external parsers.
 * @param String[] args - command line arguments (as in main)
*/
	public MX(String[] args) {
		CmdLineArgs cla = util.getCmdLine(args);
		parserName = cla.parserName;
		inFile = cla.inFile;
		outFile = cla.outFile;
		debugFile = cla.debugFile;
		verbose = cla.verbose;
		init();
		varInit();
		if (verbose) {
			verboseInit();
		}
	}

/**
 * convert a midi file to an XMidi (XML) file
*/
	public void midiToXMidi() {
		varInit();
		if (debugFile != null) {
			try {
				deb = new FileWriter(debugFile);
			} catch(IOException e) {
				util.ept(e,"Unable to open "+debugFile);
			}
		}

		Document xDoc = midiToDoc(inFile);
		docToXMidi(xDoc, outFile);

		if (deb != null) {
			try {
				deb.close();
			} catch(IOException e) {
				util.ept(e,"Unable to close "+debugFile);
			}
		}
	}

	void closeDeb() {
		if (deb != null) {
			try {
				deb.close();
			} catch(IOException e) {
				util.ept(e,"Unable to close "+debugFile);
			}
		}
	}

/**
 * Convert a MIDI file to a DOM structure.
 * @return Document
*/
	public Document midiToDoc(String inFile) {
		try {
			fis = new FileInputStream(inFile);
		} catch(FileNotFoundException e) {
			util.ept(e,"Unable to open "+inFile);
		}
		Document xDoc = util.getDoc(inFile, false, parserName);
		doDoc(xDoc);

		try {
			fis.close();
		} catch(IOException e) {
			util.ept(e,"Unable to close "+inFile);
		}

		return xDoc;
	}

/**
 * Convert a DOM structure to an XMidi file.
 * This ought to be part of the DOM interface,
 * but isn't, so I have to implement it here.
 * @param Document xDoc - the &quot;base&quot; of the DOM structure
 * @param String outFile - the file name of the file to write
*/
	public void docToXMidi(Document xDoc, String outFile) {
		try {
			fw = new FileWriter(outFile);
		} catch(IOException e) {
			util.ept(e,"Unable to open "+outFile);
		}
//		  wpt(xDoc.toString());
		doHeader();
		Element root = xDoc.getDocumentElement();
spt("--- about to cal doElem with root");
		doElem(root, 0);
		try {
			fw.close();
		} catch(IOException e) {
			util.ept(e,"Error closing " + outFile);
		}
	}

	void doElem(Element elem, int rnum) {
		if (elem == null) {
spt("--- null element");
			closeDeb();
			util.rpt("null Element in doElem");
		}
		String tag = elem.getTagName();
spt("--- in doElem: Tag is "+tag);
		String o = "<" + tag;
		NamedNodeMap nnm = elem.getAttributes();
		if (nnm != null) {
			// for each attribute
			int n = nnm.getLength();
			for (int i = 0; i < n; i ++) {
				Attr attr = (Attr)nnm.item(i);
				String aNam = attr.getName();
				String aVal = attr.getValue();
				o += " "+aNam+"=";
				if (aVal.indexOf("\"") == -1) {
					o += "\""+aVal+"\"";
				} else if (aVal.indexOf("'") == -1) {
					o += "'"+aVal+"'";
				} else {
					util.rpt("Invalid value " + "("+aVal+")"+
						" for attribute "+aNam+" in tag "+tag);
				}
			}
		}
		NodeList nl = elem.getChildNodes();
		int inl = nl.getLength();
		if (inl == 0) {
			o += "/>";
		} else {
			o += ">";
		}
		String indent = "";
		for (int i = 0; i < rnum; i ++) {
			indent += "    ";
		}
/*
		if (tag.equals("EDATA")) {
			String oo = indent + o;
			// for each child
			for (int i = 0; i < inl; i ++) {
				Node n = nl.item(i);
				short nt = n.getNodeType();
				if (nt == Node.ELEMENT_NODE) {
					util.rpt("doing EDATA; there should be no Element kids: "+
						n.getNodeName());
//					  doElem((Element)n, rnum + 1);
				} else if (nt == Node.TEXT_NODE) {
					Text tnt = (Text)n;
					oo += tnt.getData().trim();
//					  wpt(indent + "    " + tnt.getData());
				} else {
					util.rpt("node type is "+nt);
				}
			}
			if (inl != 0) {
				oo += "</" + tag + ">";
//				  wpt(indent + "</" + tag + ">");
			}
			wpt(oo);
		} else {
*/
			wpt(indent + o);

			// for each child
			for (int i = 0; i < inl; i ++) {
				Node n = nl.item(i);
				short nt = n.getNodeType();
				if (nt == Node.ELEMENT_NODE) {
					doElem((Element)n, rnum + 1);
				} else if (nt == Node.TEXT_NODE) {
					Text tnt = (Text)n;
					wpt(indent + "    " + tnt.getData());
				} else {
					util.rpt("node type is "+nt);
				}
			}
			if (inl != 0) {
				wpt(indent + "</" + tag + ">");
			}
//		  }   // end else (= EDATA)
	}

	void doHeader() {
		// write the header
		wpt("<?xml version=\"1.0\" standalone=\"no\"?>");
		wpt("<!DOCTYPE XMidi SYSTEM \"XMidi.dtd\">");
		wpt("<!-- produced by com.palserv.XMidi.MX from "+inFile+" -->");
	}

	Element addElem(Document xDoc, Element parent, String tag) {

		if (parent == null) {
			util.rpt("addElem: null parent adding "+tag);
		}
		Element child = null;
		try {
			child = xDoc.createElement(tag);
		} catch(DOMException de) {
			util.ept(de,"DOMException creating "+tag+" element");
		}

		try {
			parent.appendChild(child);
		} catch(DOMException de) {
			util.ept(de,"DOMException appending "+tag+" to "+
				parent.getTagName());
		}
		return child;
	}

	void addAttr(Element elem, String name, String val) {
		try {
			elem.setAttribute(name, val);
		} catch(DOMException de) {
			util.ept(de,"DOMException setting "+name+" attribute to " +
				val + " on "+elem.getTagName()+ " element");
		}
	}

	void addText(Document xDoc, Element elem, String txt) {
		Text newText = null;
		try {
			newText = xDoc.createTextNode(txt);
		} catch(DOMException de) {
			util.ept(de,"DOMException creating text element for "+
				elem.getTagName());
		}

		try {
			elem.appendChild(newText);
		} catch(DOMException de) {
			util.ept(de,"DOMException appending text to " +
				elem.getTagName());
		}
	}

	void doDoc(Document xDoc) {
		Element root = null;
		try {
			root = xDoc.createElement("XMidi");
		} catch(DOMException de) {
			util.ept(de,"DOMException creating xml processing instruction");
		}

		try {
			xDoc.appendChild(root);
		} catch(DOMException de) {
			util.ept(de,"DOMException appending root to document");
		}

		addAttr(root,"VERSION", Version._VERSION);

		boolean cflag = true;
		while (cflag) { 	// for each chunk
			byte[] hdr = new byte[8];
			try {
				int r = fis.read(hdr);
				if (r == -1) {
					cflag = false;
					continue;
				}
				if (r < 8) {
					util.rpt("getting header");
				}
			} catch(IOException e) {
				util.ept(e,"IO exception getting header");
			}
			// get chunk id
			String cid = baToString(hdr,0,3);

			// get chunk length
			int len = util.baToInt(hdr,4,7);

			// get rest of chunk
			byte[] dta = new byte[len];
			try {
				int r = fis.read(dta);
				if (r < len) {
					util.rpt("getting data");
				}
			} catch(IOException e) {
				util.ept(e,"IO exception getting data");
			}

			if (cid.equals("MThd")) {
				Element mthd = addElem(xDoc, root, "MThd");
				addAttr(mthd,"TYPE", cid);
				addAttr(mthd,"LENGTH", ""+len);

				midiFormat = util.baToInt(dta,0,1);
				numTracks = util.baToInt(dta,2,3);
				ppnq = util.baToInt(dta,4,5);
				String pnq = util.baToHex(dta,4,5);

				Element eFormat = addElem(xDoc, mthd, "FORMAT");
				addText(xDoc, eFormat, ""+midiFormat);

				Element eTracks = addElem(xDoc, mthd, "TRACKS");
				addText(xDoc, eTracks, ""+numTracks);

				Element ePnq = addElem(xDoc, mthd, "PPNQ");
				addText(xDoc, ePnq, pnq);

			} else if (cid.equals("MTrk")) {
				trkCount ++;
				if (trkCount > numTracks) {
					util.rpt("too many tracks");
				}

				Element mtrk = addElem(xDoc, root, "MTrk");
				addAttr(mtrk, "TYPE", cid);
				addAttr(mtrk, "LENGTH", ""+len);
				addAttr(mtrk, "NUMBER", ""+trkCount);

				doTrack(dta,len, mtrk, xDoc);
			} else {
				Element chnk = addElem(xDoc, root, "CHUNK");
				addAttr(chnk, "TYPE", cid);
				addAttr(chnk, "LENGTH", ""+len);

				doHexData(dta,len,fw, chnk, xDoc);
			}

		}	// end of for each chunk
		spt("Finished");
	}

/**
 * main routine, executed from command line.
 * See the
 * {@link com.palserv.XMidi.Util#getCmdLine(String[]) Util.getCmdLine}
 * method for details of the command line arguments
*/
	public static void main(String[] args) {
		MX mx = new MX(args);
		mx.midiToXMidi();
	}

	void varInit() {
		fw = null;
		fis = null;
		numTracks = 0;
		trkCount = 0;
	}

	void init() {
		ffHash = new Hashtable<String,String>();
		ffHash.put("00", "Sequence Number");
		ffHash.put("01", "Text");
		ffHash.put("02", "Copyright");
		ffHash.put("03", "Sequence/Track Name");
		ffHash.put("04", "Instrument");
		ffHash.put("05", "Lyric");
		ffHash.put("06", "Marker");
		ffHash.put("07", "Cue Point");
		ffHash.put("20", "MIDI Channel");
		ffHash.put("21", "MIDI Port");
		ffHash.put("2F", "End of Track");
		ffHash.put("51", "Tempo");
		ffHash.put("54", "SMPTE Offset");
		ffHash.put("58", "Time Signature");
		ffHash.put("59", "Key Signature");
		ffHash.put("7F", "Proprietary Event");

		chanArray = new String[7];
		chanArray[0] = "Note Off";
		chanArray[1] = "Note On";
		chanArray[2] = "After Touch";
		chanArray[3] = "Control Change";
		chanArray[4] = "Program Change";
		chanArray[5] = "Channel Pressure";
		chanArray[6] = "Pitch Wheel";

		fArray = new String[16];
		fArray[0] = "SYSEX";
		fArray[1] = "MTC Quarter Frame Message";
		fArray[2] = "Song Position Pointer";
		fArray[3] = "Song Select";
		fArray[4] = "Undefined";
		fArray[5] = "Undefined";
		fArray[6] = "Tune Request";
		fArray[7] = "Unsupported";
		fArray[8] = "MIDI Clock";
		fArray[9] = "Undefined";
		fArray[10] = "MIDI Start";
		fArray[11] = "MIDI Continue";
		fArray[12] = "MIDI Stop";
		fArray[13] = "Undefined";
		fArray[14] = "Active Sense";
		fArray[15] = "NotUnderstood";
	}

	void verboseInit() {
		contHash = new Hashtable<String,String>();

		contHash.put("0", "Bank Select");
		contHash.put("1", "Modulation Wheel (coarse)");
		contHash.put("2", "Breath controller (coarse)");
		contHash.put("4", "Foot Pedal (coarse)");
		contHash.put("5", "Portamento Time (coarse)");
		contHash.put("6", "Data Entry (coarse)");
		contHash.put("7", "Volume (coarse)");
		contHash.put("8", "Balance (coarse)");
		contHash.put("10", "Pan position (coarse)");
		contHash.put("11", "Expression (coarse)");
		contHash.put("12", "Effect Control 1 (coarse)");
		contHash.put("13", "Effect Control 2 (coarse)");
		contHash.put("16", "General Purpose Slider 1");
		contHash.put("17", "General Purpose Slider 2");
		contHash.put("18", "General Purpose Slider 3");
		contHash.put("19", "General Purpose Slider 4");
		contHash.put("32", "Bank Select (fine)");
		contHash.put("33", "Modulation Wheel (fine)");
		contHash.put("34", "Breath controller (fine)");
		contHash.put("36", "Foot Pedal (fine)");
		contHash.put("37", "Portamento Time (fine)");
		contHash.put("38", "Data Entry (fine)");
		contHash.put("39", "Volume (fine)");
		contHash.put("40", "Balance (fine)");
		contHash.put("42", "Pan position (fine)");
		contHash.put("43", "Expression (fine)");
		contHash.put("44", "Effect Control 1 (fine)");
		contHash.put("45", "Effect Control 2 (fine)");
		contHash.put("64", "Hold Pedal (on/off)");
		contHash.put("65", "Portamento (on/off)");
		contHash.put("66", "Sustenuto Pedal (on/off)");
		contHash.put("67", "Soft Pedal (on/off)");
		contHash.put("68", "Legato Pedal (on/off)");
		contHash.put("69", "Hold 2 Pedal (on/off)");
		contHash.put("70", "Sound Variation");
		contHash.put("71", "Sound Timbre");
		contHash.put("72", "Sound Release Time");
		contHash.put("73", "Sound Attack Time");
		contHash.put("74", "Sound Brightness");
		contHash.put("75", "Sound Control 6");
		contHash.put("76", "Sound Control 7");
		contHash.put("77", "Sound Control 8");
		contHash.put("78", "Sound Control 9");
		contHash.put("79", "Sound Control 10");
		contHash.put("80", "General Purpose Button 1 (on/off)");
		contHash.put("81", "General Purpose Button 2 (on/off)");
		contHash.put("82", "General Purpose Button 3 (on/off)");
		contHash.put("83", "General Purpose Button 4 (on/off)");
		contHash.put("91", "Effects Level");
		contHash.put("92", "Tremulo Level");
		contHash.put("93", "Chorus Level");
		contHash.put("94", "Celeste Level");
		contHash.put("95", "Phaser Level");
		contHash.put("96", "Data Button increment");
		contHash.put("97", "Data Button decrement");
		contHash.put("98", "Non-registered Parameter (fine)");
		contHash.put("99", "Non-registered Parameter (coarse)");
		contHash.put("100", "Registered Parameter (fine)");
		contHash.put("101", "Registered Parameter (coarse)");
		contHash.put("120", "All Sound Off");
		contHash.put("121", "All Controllers Off");
		contHash.put("122", "Local Keyboard (on/off)");
		contHash.put("123", "All Notes Off");
		contHash.put("124", "Omni Mode Off");
		contHash.put("125", "Omni Mode On");
		contHash.put("126", "Mono Operation");
		contHash.put("127", "Poly Operation");

		notes = new String[128];
		register = new int[128];
		notes[0] = "C";
		register[0] = -5;
		notes[1] = "C#";
		register[1] = -5;
		notes[2] = "D";
		register[2] = -5;
		notes[3] = "Eb";
		register[3] = -5;
		notes[4] = "E";
		register[4] = -5;
		notes[5] = "F";
		register[5] = -5;
		notes[6] = "F#";
		register[6] = -5;
		notes[7] = "G";
		register[7] = -5;
		notes[8] = "Ab";
		register[8] = -5;
		notes[9] = "A";
		register[9] = -5;
		notes[10] = "Bb";
		register[10] = -5;
		notes[11] = "B";
		register[11] = -5;
		notes[12] = "C";
		register[12] = -4;
		notes[13] = "C#";
		register[13] = -4;
		notes[14] = "D";
		register[14] = -4;
		notes[15] = "Eb";
		register[15] = -4;
		notes[16] = "E";
		register[16] = -4;
		notes[17] = "F";
		register[17] = -4;
		notes[18] = "F#";
		register[18] = -4;
		notes[19] = "G";
		register[19] = -4;
		notes[20] = "Ab";
		register[20] = -4;
		notes[21] = "A";
		register[21] = -4;
		notes[22] = "Bb";
		register[22] = -4;
		notes[23] = "B";
		register[23] = -4;
		notes[24] = "C";
		register[24] = -3;
		notes[25] = "C#";
		register[25] = -3;
		notes[26] = "D";
		register[26] = -3;
		notes[27] = "Eb";
		register[27] = -3;
		notes[28] = "E";
		register[28] = -3;
		notes[29] = "F";
		register[29] = -3;
		notes[30] = "F#";
		register[30] = -3;
		notes[31] = "G";
		register[31] = -3;
		notes[32] = "Ab";
		register[32] = -3;
		notes[33] = "A";
		register[33] = -3;
		notes[34] = "Bb";
		register[34] = -3;
		notes[35] = "B";
		register[35] = -3;
		notes[36] = "C";
		register[36] = -2;
		notes[37] = "C#";
		register[37] = -2;
		notes[38] = "D";
		register[38] = -2;
		notes[39] = "Eb";
		register[39] = -2;
		notes[40] = "E";
		register[40] = -2;
		notes[41] = "F";
		register[41] = -2;
		notes[42] = "F#";
		register[42] = -2;
		notes[43] = "G";
		register[43] = -2;
		notes[44] = "Ab";
		register[44] = -2;
		notes[45] = "A";
		register[45] = -2;
		notes[46] = "Bb";
		register[46] = -2;
		notes[47] = "B";
		register[47] = -2;
		notes[48] = "C";
		register[48] = -1;
		notes[49] = "C#";
		register[49] = -1;
		notes[50] = "D";
		register[50] = -1;
		notes[51] = "Eb";
		register[51] = -1;
		notes[52] = "E";
		register[52] = -1;
		notes[53] = "F";
		register[53] = -1;
		notes[54] = "F#";
		register[54] = -1;
		notes[55] = "G";
		register[55] = -1;
		notes[56] = "Ab";
		register[56] = -1;
		notes[57] = "A";
		register[57] = -1;
		notes[58] = "Bb";
		register[58] = -1;
		notes[59] = "B";
		register[59] = -1;
		notes[60] = "C";
		register[60] = 0;
		notes[61] = "C#";
		register[61] = 0;
		notes[62] = "D";
		register[62] = 0;
		notes[63] = "Eb";
		register[63] = 0;
		notes[64] = "E";
		register[64] = 0;
		notes[65] = "F";
		register[65] = 0;
		notes[66] = "F#";
		register[66] = 0;
		notes[67] = "G";
		register[67] = 0;
		notes[68] = "Ab";
		register[68] = 0;
		notes[69] = "A";
		register[69] = 0;
		notes[70] = "Bb";
		register[70] = 0;
		notes[71] = "B";
		register[71] = 0;
		notes[72] = "C";
		register[72] = 1;
		notes[73] = "C#";
		register[73] = 1;
		notes[74] = "D";
		register[74] = 1;
		notes[75] = "Eb";
		register[75] = 1;
		notes[76] = "E";
		register[76] = 1;
		notes[77] = "F";
		register[77] = 1;
		notes[78] = "F#";
		register[78] = 1;
		notes[79] = "G";
		register[79] = 1;
		notes[80] = "Ab";
		register[80] = 1;
		notes[81] = "A";
		register[81] = 1;
		notes[82] = "Bb";
		register[82] = 1;
		notes[83] = "B";
		register[83] = 1;
		notes[84] = "C";
		register[84] = 2;
		notes[85] = "C#";
		register[85] = 2;
		notes[86] = "D";
		register[86] = 2;
		notes[87] = "Eb";
		register[87] = 2;
		notes[88] = "E";
		register[88] = 2;
		notes[89] = "F";
		register[89] = 2;
		notes[90] = "F#";
		register[90] = 2;
		notes[91] = "G";
		register[91] = 2;
		notes[92] = "Ab";
		register[92] = 2;
		notes[93] = "A";
		register[93] = 2;
		notes[94] = "Bb";
		register[94] = 2;
		notes[95] = "B";
		register[95] = 2;
		notes[96] = "C";
		register[96] = 3;
		notes[97] = "C#";
		register[97] = 3;
		notes[98] = "D";
		register[98] = 3;
		notes[99] = "Eb";
		register[99] = 3;
		notes[100] = "E";
		register[100] = 3;
		notes[101] = "F";
		register[101] = 3;
		notes[102] = "F#";
		register[102] = 3;
		notes[103] = "G";
		register[103] = 3;
		notes[104] = "Ab";
		register[104] = 3;
		notes[105] = "A";
		register[105] = 3;
		notes[106] = "Bb";
		register[106] = 3;
		notes[107] = "B";
		register[107] = 3;
		notes[108] = "C";
		register[108] = 4;
		notes[109] = "C#";
		register[109] = 4;
		notes[110] = "D";
		register[110] = 4;
		notes[111] = "Eb";
		register[111] = 4;
		notes[112] = "E";
		register[112] = 4;
		notes[113] = "F";
		register[113] = 4;
		notes[114] = "F#";
		register[114] = 4;
		notes[115] = "G";
		register[115] = 4;
		notes[116] = "Ab";
		register[116] = 4;
		notes[117] = "A";
		register[117] = 4;
		notes[118] = "Bb";
		register[118] = 4;
		notes[119] = "B";
		register[119] = 4;
		notes[120] = "C";
		register[120] = 5;
		notes[121] = "C#";
		register[121] = 5;
		notes[122] = "D";
		register[122] = 5;
		notes[123] = "Eb";
		register[123] = 5;
		notes[124] = "E";
		register[124] = 5;
		notes[125] = "F";
		register[125] = 5;
		notes[126] = "F#";
		register[126] = 5;
		notes[127] = "G";
		register[127] = 5;
	}

/**
write output line
*/
	void wpt(String s) {
		if (fw == null) {
			util.rpt("in wpt: fw is null trying to write: "+s);
		}
		spt(s);
		try {
			fw.write(s+"\n");
		} catch(IOException e) {
			util.ept(e,"Error writing to "+outFile);
		}
	}

/**
add track data to DOM structure
*/
	void doTrack(byte[] dta, int len, Element mtrk, Document xDoc) {
		boolean tFlag = true;
		int offset = 0;

		// initialize variables
		int dlen = 0;
		String edata = null;
		String snam = null;
		String nmData = "";
		String ctag = null;
		String ctagAttr[] = new String[4];
		String ctagAttrVal[] = new String[4];
		int ctagnum = 0;
		String cnum = null;
		String ctyp = null;
		int slen = 0;
		int chanType = 0;
		int hiName = 0;
		int status = 0;
		int noff = 0;
		String sval = null;
		boolean ecFlag = true; // assume edata
		boolean nFlag = true;  // use slen for noff
		boolean firstTime = true;

		while (tFlag) {
			// do delta
			ByteLen bl = deltaToInt(dta, offset);
			offset += bl.len;
			String deltaLen = util.baToHex(bl.ba,0,3);

			nFlag = true;	// assume simple (slen) offset
			// may or may not be status
			boolean statFlag = false;
			int first = util.baToInt(dta,offset,offset);
spt("doTrack: in loop; deltaLen="+deltaLen+", len="+bl.len+", first="+first);
			if ((first & 128) == 128) {
				// it is a status byte
				statFlag = true;
				sval = util.baToHex(dta,offset,offset);
				status = first;
spt("doTrack: have status: "+sval);
				if (status < 240 && status > 127) {
					ecFlag = false;
					chanType = (status - 128) / 16;
					snam = chanArray[chanType];
					ctyp = sval.substring(0,1);
					cnum = sval.substring(1,2);
				} else {
					ecFlag = true;
					if (status > 239 && status < 256) {
						hiName = status - 240;
						snam = fArray[hiName];
					} else {
						util.rpt("Invalid status: "+status);
					}
				}
				offset ++;
			} else {
spt("doTrack: running status");
			}

			nmData = "";
			if (firstTime) {
				firstTime = false;
				if (!statFlag) {
					util.rpt("first time, but no status; first = "+first);
				}
			}

			// offset points to the byte after the status
			// or first byte of "running status"
			Element delta = addElem(xDoc, mtrk, "DELTA");
			addAttr(delta, "DTIME", ""+deltaLen);

			if (status > 127 && status < 144) {
				// note off
				slen = 2;
				// set up tag
				int pitch = util.baToInt(dta,offset,offset);
				int vel = util.baToInt(dta,offset+1,offset+1);
				ctag = "NOTE_OFF";
				ctagAttr[0] = "PITCH";
				ctagAttrVal[0] = ""+pitch;
				ctagAttr[1] = "VELOCITY";
				ctagAttrVal[1] = ""+vel;
				ctagnum = 2;
				if (verbose) {
					ctagAttr[2] = "NAME";
					ctagAttrVal[2] = notes[pitch];
					ctagAttr[3] = "REGISTER";
					ctagAttrVal[3] = ""+register[pitch];
					ctagnum = 4;
				}
			} else if (status > 143 && status < 160) {
				// note on
				slen = 2;
				int pitch = util.baToInt(dta,offset,offset);
				int vel = util.baToInt(dta,offset+1,offset+1);
				ctag = "NOTE_ON";
				ctagAttr[0] = "PITCH";
				ctagAttrVal[0] = ""+pitch;
				ctagAttr[1] = "VELOCITY";
				ctagAttrVal[1] = ""+vel;
				ctagnum = 2;
				if (verbose) {
					ctagAttr[2] = "NAME";
					ctagAttrVal[2] = notes[pitch];
					ctagAttr[3] = "REGISTER";
					ctagAttrVal[3] = ""+register[pitch];
					ctagnum = 4;
				}
			} else if (status > 159 && status < 176) {
				// after touch
				slen = 2;
				int pitch = util.baToInt(dta,offset,offset);
				int pres = util.baToInt(dta,offset+1,offset+1);
				ctag = "AFTER";
				ctagAttr[0] = "PITCH";
				ctagAttrVal[0] = ""+pitch;
				ctagAttr[1] = "PRESSURE";
				ctagAttrVal[1] = ""+pres;
				ctagnum = 2;
				if (verbose) {
					ctagAttr[2] = "NAME";
					ctagAttrVal[2] = notes[pitch];
					ctagAttr[3] = "REGISTER";
					ctagAttrVal[3] = ""+register[pitch];
					ctagnum = 4;
				}
			} else if (status > 175 && status < 192) {
				// control change
				slen = 2;
				int contnum = util.baToInt(dta,offset,offset);
				int contval = util.baToInt(dta,offset+1,offset+1);
				ctag = "CONTROL";
				ctagAttr[0] = "NUMBER";
				ctagAttrVal[0] = ""+contnum;
				ctagAttr[1] = "VALUE";
				ctagAttrVal[1] = ""+contval;
				ctagnum = 2;
				if (verbose) {
					ctagAttr[2] = "NAME";
					ctagAttrVal[2] = (String)contHash.get(""+contnum);
					ctagnum = 3;
				}
			} else if (status > 191 && status < 208) {
				// program (patch) change
				slen = 1;
				int patch = util.baToInt(dta,offset,offset);
				ctag = "PROGRAM";
				ctagAttr[0] = "NUMBER";
				ctagAttrVal[0] = ""+patch;
				ctagnum = 1;
			} else if (status > 207 && status < 224) {
				// channel pressure
				slen = 1;
				int pamt = util.baToInt(dta,offset,offset);
				ctag = "PRESSURE";
				ctagAttr[0] = "AMOUNT";
				ctagAttrVal[0] = ""+pamt;
				ctagnum = 1;
			} else if (status > 223 && status < 240) {
				// pitch wheel
				slen = 2;
				int pwamt = getPW(dta,offset);
				ctag = "WHEEL";
				ctagAttr[0] = "AMOUNT";
				ctagAttrVal[0] = ""+pwamt;
				ctagnum = 1;
			} else if (status == 240) {
				// sysex
				bl = deltaToInt(dta, offset);
				slen = util.baToInt(bl.ba,0,3);
				noff = bl.len;
				nFlag = false;
				edata = util.baToHex(dta,offset+noff,offset+noff+slen-1);
				noff += slen;
			} else if (status == 255) {
				// non midi (reset only in "live" midi")
				nmData = util.baToHex(dta,offset,offset);
//				  nmData = "SNMT=\""+nmDta+"\" ";
				snam = "non-MIDI";
				String nmNam = (String)ffHash.get(nmData);
				if (nmNam != null) {
					snam += " - "+nmNam;
				} else {
					snam += " - Unknown";
				}
//				  int nmt = util.baToInt(dta,offset+1,offset+1);
				bl = deltaToInt(dta, offset + 1);
				slen = util.baToInt(bl.ba,0,3);
				noff = bl.len+1;
				nFlag = false;
				if (slen == 0) {
					edata = "No data";
				} else {
					edata = util.baToHex(dta,offset+noff,
						offset+noff+slen-1);
					noff += slen;
				}
				if (nmData.equals("58")) { // time sig
					Element timeSig = addElem(xDoc, delta, "TIME_SIG");
					int nn = util.baToInt(dta, offset + 2, offset + 2);
					int dd = util.baToInt(dta, offset + 3, offset + 3);
					int cc = util.baToInt(dta, offset + 4, offset + 4);
					int bb = util.baToInt(dta, offset + 5, offset + 5);
					int ddd = (int)Math.pow(2, dd);
					addAttr(timeSig, "NN", ""+nn);
					addAttr(timeSig, "DD", ""+dd);
					addAttr(timeSig, "CC", ""+cc);
					addAttr(timeSig, "BB", ""+bb);
					addAttr(timeSig, "SIG", ""+nn+"/"+ddd);
				}
spt("doTrack: status FF"+nmData+", edata = "+edata);
			} else if (status == 241 || status == 243) {
				int tcv = (int)dta[offset];
				Integer tc = new Integer(tcv);
				edata = tc.toString();
				slen = 1;
			} else if (status == 242) {
				int tcv = getPW(dta,offset);
				Integer tc = new Integer(tcv);
				edata = tc.toString();
				slen = 2;
			} else if (status == 246 || status == 248 ||
					status == 250 || status == 251 ||
					status == 252 || status == 254) {
				edata = "No data for "+snam;
				slen = 0;
			} else {	// really unknown
				int ol = getNextHiOrd(dta, offset);
				edata = util.baToHex(dta,offset+1,ol);
				ol -= offset + 1;
				slen = ol;
			}
			Element eStatus = null;
			if (statFlag) {
				eStatus = addElem(xDoc, delta, "STATUS");
				addAttr(eStatus, "SNAM", snam);
				addAttr(eStatus, "SVAL", sval);
				addAttr(eStatus, "SLEN", ""+slen);
				if (!nmData.equals("")) {
					addAttr(eStatus, "SNMT", nmData);
				}
			}
			if (ecFlag) {
				Element edat = null;
				if (statFlag) {
					edat = addElem(xDoc, eStatus, "EDATA");
				} else {
					edat = addElem(xDoc, delta, "EDATA");
				}
				addText(xDoc, edat, edata);
			} else {
				Element channel = null;
				if (statFlag) {
					channel = addElem(xDoc, eStatus, "CHANNEL");
				} else {
					channel = addElem(xDoc, delta, "CHANNEL");
				}
				addAttr(channel, "TYPE", ctyp);
				addAttr(channel, "NUMBER", cnum);
				Element subchan = addElem(xDoc, channel, ctag);
				for (int c = 0; c < ctagnum; c ++) {
					addAttr(subchan, ctagAttr[c], ctagAttrVal[c]);
				}
			}

			if (nFlag) {
				offset += slen;
			} else {
				offset += noff;
			}
			if (offset >= len) {
				tFlag = false;
			}
		}
	}

	int getPW(byte[] dta, int offset) {
		int hi = util.baToInt(dta,offset,offset);
		int lo = util.baToInt(dta,offset+1,offset+1);
		hi <<= 7;
		lo |= hi;
		return lo;
	}

	int getNextHiOrd(byte[] dta, int offset) {
		int ol = 0;
		boolean tflag = true;
		for (int o = offset+1; o < dta.length - 1; o ++) {
			if (tflag) {
				int x = util.baToInt(dta,o,o);
				if ((x & 128) == 128) {
					tflag = false;
					ol = o;
				}
			}
		}
		if (tflag) {
			ol = offset+dta.length;
		}
		return ol;
	}

/**
write formatted hex data to file
*/
	void doHexData(byte[] dta, int len, FileWriter fw,
			Element chnk, Document xDoc) {
//		  wpt("        <HEXDATA>");
		Element hdata = addElem(xDoc, chnk, "HEXDATA");
		StringBuffer bth = new StringBuffer("");
		int bpl = 32;
		int r = len % bpl;
		int n = len / bpl;
		for (int i = 0; i < n; i ++) {
			int strt = i * bpl;
//			  wpt("            "+util.baToHex(dta,strt,strt + bpl - 1));
			bth.append(util.baToHex(dta,strt,strt + bpl - 1));
		}
		if (r > 0) {
			int strt = n * bpl;
//			  wpt("            "+util.baToHex(dta,strt,strt + r - 1));
			bth.append(util.baToHex(dta,strt,strt + r - 1));
		}
		addText(xDoc, hdata, bth.toString());
//		  wpt("        </HEXDATA>");
	}

/**
write a line to standard output.
This is mostly for debugging.
for implementation, it should be a
null routine.
*/
	void spt(String s) {
//		  System.out.println(s);
		if (deb != null) {
			try {
				deb.write(s+"\n");
			} catch(IOException e) {
				util.ept(e,"error in debug write");
			}
		}
	}

/**
convert byte array to string (not hex)
*/
	String baToString(byte[] b, int strt, int end) {
		int l = end - strt + 1;
		char[] c = new char[l];
		int i = 0;
		for (int j = 0; j < l; j ++) {
			c[j] = (char)b[j];
		}
		return new String(c);
	}

/**
convert delta time to byte array
a delta time is expressed as a
byte array, length unknown (4 or less)
*/
	public ByteLen deltaToInt(byte[] b, int offset) {
/*
- capture up to four bytes including first with hi-ord
	bit off
- turn off hi-ord bits
- accumulate 4 groups of 7 into 28 bit number with
	hi-ord 4 bits zero.
*/
		int j = 0;
		byte[] ba = new byte[4];
		boolean jFlag = true;
		while (jFlag) {
			if ((j + offset) > b.length) {
				util.rpt("deltaToInt: length error; j = "+j+", offset = "+offset+", b.length = "+b.length);
			}
			ba[j] = b[j + offset];
			if (ba[j] >= 0) {
				jFlag = false;
			} else {
				ba[j] &= 127;
			}
spt("deltaToInt: j = "+j+", ba = "+util.baToInt(ba,0,j));
			j ++;
		}
		int s = 0;
		for (int i = 0; i < j; i ++) {
			int k = j - i - 1;
			int p = util.pow(128,i);
			int m = (int)ba[k];
			s += m * p;
spt("deltaToInt: in loop: s = "+s+", i = "+i+", j = "+j+", k = "+k+", p = "+p+", m = "+m);
		}
spt("deltaToInt: s = "+s);
		ByteLen bl = new ByteLen(util.intToBa(s,4),j);
		return bl;
	}
}
