/*
The following licence has been copied from
http://www.opensource.org/licenses/mit-license.html

Copyright (c) 2005, Peter Arthur Loeb

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package com.palserv.XMidi;

import org.w3c.dom.*;
import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import java.io.*;
import java.util.*;

/**
 * Convert an XMidi file (XML) to a midi file.
 * <p>
 * This class can be used in two ways:
 * <ol>
 * <li>
 * From the command line.
 * <li>
 * From another java class:
 * <ol>
 * <li>
 * Create an array of Strings
 * which will correspond to the
 * command line arguments.
 * <li>
 * Pass the array to the constructor:
 * <pre>
 * XM xm = new XM(args);
 * </pre>
 * <li>
 * Invoke the main conversion method:
 * <pre>
 * xm.xMidiToMidi();
 * </pre>
 * </ol>
 * </ol>
 * <p>
 * Even though the XMidi format has a DTD (XMidi.dtd)
 * associated with it, this program does quite a bit of
 * its own validation.	This may be somewhat redundant,
 * but it will allow the program to work on an XMidi file
 * which does not name a DTD.  The validation forces many
 * of the same things as the DTD, although there are differences.
 * @author Peter Arthur Loeb
 * <p>
 * Questions or comments should be addressed to the author
 * by email: peter@palserv.com
*/
public class XM extends DefaultHandler {
	private String inFile;
	private String outFile;
	private PrintWriter pw;
	private State state;

/**
 * public constructor
 *
 * Takes &quot;command line&quot; arguments.
 * see the
 * {@link com.palserv.XMidi.Util#getCmdLine(String[]) Util.getCmdLine}
 * method for details.
 * Also, see the main package description
 * for details about external parsers.
 * @param args - command line arguments (as in main)
 */
	public XM(String[] args) {
		CmdLineArgs cla = Util.getCmdLine(args);
		inFile = cla.inFile;
		outFile = cla.outFile;
	}

/**
 * Convert an XMidi file (XML) to a midi file.
 * Use a SAX parser for speed.
 */
	public void xMidiToMidi() {
		SAXParserFactory spf = null;
		try {
			spf = SAXParserFactory.newInstance();
		} catch(FactoryConfigurationError e) {
			Util.rpt("FactoryConfigurationError trying to get factory: "+e.getMessage());
		}

		SAXParser sp = null;
		try {
			sp = spf.newSAXParser();
		} catch(ParserConfigurationException e) {
			Util.rpt("ParserConfigurationException trying to get parser: "+e.getMessage());
		} catch(SAXException e) {
			Util.rpt("SAXException trying to get parser: "+e.getMessage());
		}

		state = new State();
		state.chunkFlag = false;
		state.edFlag = false;
		state.parents = new Stack<String>();

		try {
			state.fos = new FileOutputStream(outFile);
		} catch(FileNotFoundException e) {
			Util.ept(e,"unable to open "+outFile);
		}

		try {
			sp.parse(new File(inFile), this);
		} catch(SAXException e) {
			Util.rpt("SAXException trying to parse: "+e.getMessage());
		} catch(IOException e) {
			Util.rpt("IOException trying to parse: "+e.getMessage());
		}

		try {
			state.fos.close();
		} catch(IOException e) {
			Util.ept(e,"error closing "+outFile);
		}
	}

	void spt(String s) {
		System.out.println(s);
	}

/**
 * Process each (XML) element when it is encountered.
 */
	public void startElement(String uri, String localName, String qName, Attributes attrs) {
//spt("starting element "+qName);
		// Get the Tag subclass which handles this element.
		Tag tag = Tag.getTag(qName);
		// Invoke the startElement method on the tag subclass
		tag.startElement(attrs, state);
		// push this element name onto the stack to keep track of parents
		state.parents.push(qName);
	}

/**
 * Process the end of each (XML) element.
 */
	public void endElement(String uri, String localName, String qName) {
//spt("ending element "+qName);
		// Get the Tag subclass which handles this element.
		Tag tag = Tag.getTag(qName);
		// Invoke the endElement method on the tag subclass
		tag.endElement(state);
		// pop this element off stack
		state.parents.pop();
	}

/**
 * Process each element's contents.
 */
	public void characters(char[] ch, int start, int length) {
		// convert to String
		String t = new String(ch, start, length);
		// get rid of white space
		// (this is what String.trim() is supposed to do, but doesn't)
//		  s = Util.fixString(s);
		String s = t.trim();
		if (s == null || s.equals("")) {
			return;
		}
		// Get the Tag subclass which handles this element.
		Tag tag = Tag.getTag(state.getParent());
		// invoke the doContent method to handle content
//Util.spt("XM.characters: invoking doContent on "+state.getParent()+", content = "+s);
		tag.doContent(s, state);
	}

/**
 * This method is for command line invocation.
 * See the
 * {@link com.palserv.XMidi.Util#getCmdLine(String[]) Util.getCmdLine}
 * method for details of the command line syntax.
 * @param args - command line
*/
	public static void main(String[] args) {
		XM xm = new XM(args);
		xm.xMidiToMidi();
	}
}
